"""
Copyright (c) 2024 laffra - All Rights Reserved. 

This module provides functions for creating and managing the application menu.

"""

import logging

import ltk
import state
import constants

logger = logging.getLogger('root')


def create_menu():
    """
    Creates the application menu with the following options:
    
    - File menu:
        - New: Creates a new sheet
        - Open: Goes to the home page
        - Delete: Deletes the current sheet
    
    - View menu:
        - Full Screen: Toggles full screen mode
    
    - Help menu:
        - About: Opens the PySheets website
        - Feedback: Opens a feedback form
        - Discord: Opens the PySheets Discord server
    
    The menu is animated to fade in when it is created.
    """
    def go_home():
        ltk.window.document.location = "/"

    ltk.find(".logo").on("click", ltk.proxy(lambda event: go_home()))

    return ltk.MenuBar([
        create_file_menu(),
        create_view_menu(),
        create_help_menu()
    ]).css("opacity", 0).animate(ltk.to_js({ "opacity": 1 }), constants.ANIMATION_DURATION)


def handle_import_preview(data):
    ltk.find("#import-data-preview").html(data["preview"])
    ltk.find(".import-final-button").attr("disabled", False)


def handle_import_done(data):
    ltk.find("#import-dialog").remove()


ltk.subscribe("Importer", constants.TOPIC_WORKER_PREVIEW_IMPORTED_WEB, handle_import_preview)
ltk.subscribe("Importer", constants.TOPIC_WORKER_IMPORTED_WEB, handle_import_done)


def import_sheet():
    """
    Show a dialog to let the user import a sheet.
    """

    if state.UI.editor.get():
        return ltk.window.alert("Please select an empty cell and press 'import' again.")

    def load_from_web(event):
        ltk.publish(
            "Importer",
            "Worker",
            constants.TOPIC_WORKER_PREVIEW_IMPORT_WEB,
            {
                "url": ltk.find("#import-web-url").val()
            },
        )
        # result will arrive in handle_import

    def web_url_changed(event):
        ltk.find("#import-web-url-load-button").attr("disabled", False)

    def import_into_sheet(event):
        ltk.publish(
            "Application",
            "Worker",
            constants.TOPIC_WORKER_IMPORT_WEB,
            {
                "url": ltk.find("#import-web-url").val(),
                "start_key": state.UI.current.model.key,
            },
        )
        # result will arrive in handle_import_done

    def load_dataframe(event):
        url = ltk.find("#import-web-url").val()
        state.UI.current.set(f"""=
pysheets.load_sheet("{url}")
        """)
        ltk.find("#import-dialog").remove()

    def upload_file(event):
        ltk.publish(
            "Application",
            "Worker",
            constants.TOPIC_WORKER_UPLOAD,
            {
                "id": "import-file",
            },
        )

    (ltk.Div(
        ltk.VBox(
            ltk.HBox(
                ltk.Input("")
                    .addClass("import-web-url-input")
                    .attr("id", "import-web-url")
                    .on("input", ltk.proxy(load_from_web))
                    .attr("placeholder", "Enter a web url..."),
            ),
            ltk.HBox(
                ltk.Div()
                    .addClass("import-data-preview")
                    .attr("id", "import-data-preview")
            ),
            ltk.HBox(
                ltk.Button("Import into Sheet", import_into_sheet)
                    .addClass("import-into-sheet-button")
                    .addClass("import-final-button")
                    .attr("disabled", True),
                ltk.Button("Load as Dataframe", load_dataframe)
                    .addClass("import-load-dataframe-button")
                    .addClass("import-final-button")
                    .attr("disabled", True),
                ltk.Button("Cancel", handle_import_done)
                    .addClass("import-cancel-button")
                    .addClass("import-final-button")
            )
        )
    )
    .addClass("import-dialog")
    .attr("id", "import-dialog")
    .attr("title", "Import Data")
    .dialog({
        "modal": True,
        "width": 530,
        "height": "auto"
    }))


def create_file_menu():
    """
    Create a file menu.
    """
    def go_home(event): # pylint: disable=unused-argument
        ltk.window.document.location = "/"

    def delete_sheet(event): # pylint: disable=unused-argument
        if ltk.window.confirm("This will permanently delete the current sheet."):
            import storage # pylint: disable=import-outside-toplevel
            storage.delete(
                state.UID,
                ltk.proxy(go_home),
                ltk.window.alert
            )

    items = [
        ltk.MenuItem("➕", "New", "", ltk.proxy(new_sheet)),
        ltk.MenuItem("📂", "Open", "Cmd+O", ltk.proxy(go_home)),
    ] + ([
        ltk.MenuItem("📥", "Import ...", "", ltk.proxy(lambda event: import_sheet())),
        ltk.MenuItem("🗑️", "Delete", "", ltk.proxy(delete_sheet)),
    ] if state.UID else [])
    return ltk.Menu("File", items)


def create_view_menu():
    """
    Create a view menu.
    """
    return ltk.Menu("View",
        ltk.MenuItem("⌞⌝", "Full Screen", "", lambda event: ltk.document.body.requestFullscreen()),
    )


def create_help_menu():
    """
    Create a help menu.
    """
    def about(event): # pylint: disable=unused-argument
        ltk.window.open("/about")

    def feedback(event): # pylint: disable=unused-argument
        ltk.window.open("https://docs.google.com/forms/d/e/1FAIpQLScmeDuDr5fxKYhe04Jo"
                    "-pNS73P4VF2m-i8X8EC9rfKl-jT84A/viewform")

    def discord(event): # pylint: disable=unused-argument
        ltk.window.open("https://discord.gg/4wy23872th")

    return ltk.Menu("Help",
        ltk.MenuItem("🅿️", "About", "", ltk.proxy(about)),
        ltk.MenuItem("👏", "Feedback", "", ltk.proxy(feedback)),
        ltk.MenuItem("💬", "Discord", "", ltk.proxy(discord)),
    )


def new_sheet():
    """
    Creates a new sheet and saves it to storage, then loads the new sheet.
    """
    import models # pylint: disable=import-outside-toplevel
    import storage # pylint: disable=import-outside-toplevel
    sheet = models.Sheet(uid=ltk.window.crypto.randomUUID())
    storage.save(sheet)
    load_doc(sheet.uid)


def load_doc(uid):
    """
    Loads a document with the given unique identifier (uid).
    
    Args:
        uid (str): The unique identifier of the document to load.
    
    Returns:
        None
    """
    ltk.window.document.location = f"?{constants.SHEET_ID}={uid}"



"""
Copyright (c) 2024 laffra - All Rights Reserved. 

A wrapper for the CodeMirror editor.
"""

import ltk
import lsp


class Editor(ltk.Div):
    """
    A wrapper for the CodeMirror editor that provides a convenient interface for interacting with a code editor.
    
    The `Editor` class is responsible for creating and managing a CodeMirror editor instance. It provides
    methods for setting the editor's value, getting the current cursor position, focusing
    the editor, refreshing the editor, and handling code completion.
    
    The class also sets up event listeners for the "blur" and "change" events, which 
    trigger the "change" event and clear any marks on the editor, respectively.
    """

    classes = [ "editor" ] # The CSS classes to apply to this ltk.Widget subclass

    def __init__(self, value=""):
        ltk.Div.__init__(self)
        self.editor = None
        self.code_completor = None
        self.set(value)
        self.element.css("height", "100vh")
        ltk.schedule(self.refresh, "force editor redraw", 0.5)

    def create_editor(self):
        """
        Creates a CodeMirror editor instance and sets up event listeners for the editor.
        """
        if self.editor is None:
            self.editor = ltk.window.create_codemirror_editor(self.element[0], ltk.to_js({
                "mode": {
                    "name": "python",
                    "version": 3,
                    "singleLineStringErrors": False
                },
                "lineNumbers": True,
                "indentUnit": 4,
                "extraKeys": {
                    "Ctrl-Space": "autocomplete"
                },
                "matchBrackets": True,
            }))
            self.editor.setSize("100%", "100%")
            self.editor.on("blur", ltk.proxy(lambda *args: self.trigger("change")))
            self.editor.on("change", ltk.proxy(lambda *args: self.clear_mark()))
            self.code_completor = lsp.CodeCompletor(self.editor)

    def get(self):
        """
        Returns the current value of the CodeMirror editor instance.
        """
        return self.editor.getValue()

    def get_cursor(self):
        """
        Returns the current cursor position of the CodeMirror editor instance.
        """
        return self.editor.getCursor()

    def set(self, value):
        """
        Sets the value of the CodeMirror editor instance.
        
        Args:
            value (str): The new value to set for the editor.
        """
        if self.editor and self.editor.hasFocus():
            return self
        self.create_editor()
        self.editor.setValue(str(value))
        return self

    def focus(self):
        """
        Focuses the CodeMirror editor instance. Browser keyboard events will be sent to the editor.
        
        Returns:
            self: The current instance of the editor object.
        """
        self.editor.focus()
        return self

    def refresh(self):
        """
        Refreshes the CodeMirror editor instance. This is needed to redraw the
        editor after the window has been resized.
        
        Returns:
            self: The current instance of the editor object.
        """
        self.editor.refresh()
        return self

    def handle_code_completion(self, completions):
        """
        Handles the code completion functionality for the editor computed by the worker.
        
        Args:
            completions (list): A list of code completion suggestions.
        """
        if self.code_completor:
            self.code_completor.handle_code_completion(completions)

    def clear_mark(self):
        """
        Clears the current line marker from the editor. Used to show the location of syntax errors.
        """
        ltk.window.editorClearLine()

    def mark_line(self, lineno):
        """
        Marks the specified line number in the editor with a visual indicator. 
        Used to show the location of syntax errors.
        
        Args:
            lineno (int): The line number to mark, starting from 1.
        """
        ltk.window.editorMarkLine(lineno - 1)

    def start_running(self):
        """
        Sets the editor in "run" mode.
        """
        self.find(".CodeMirror-scroll").css("background", "#f2f2f2")

    def stop_running(self):
        """
        Sets the editor in "edit" mode.
        """
        self.find(".CodeMirror-scroll").css("background", "#FFF")



"""
CopyRight (c) 2024 - Chris Laffra - All Rights Reserved.

Wrapper for OpenAI's GPT API.
"""

import json
import os
import requests

import openai


MISSING_KEY = "\n".join([
    "# We could not find an OpenAI key. Please create a file named '~/openai.json' containing:",
    "",
    "{",
    "    \"api_key\": \"YOUR_SECRET_KEY\"",
    "}",
    "",
    "# Place the file next to 'main.py' and try an OpenAI completion again.",
])


def load_key():
    """
    Loads the OpenAI API key from a JSON file located in the home directory.
    
    If the API key file is not found or there is an error reading it, the function
    will silently fail and the `openai.api_key` will remain unset.
    """
    try:
        folder = os.path.dirname(__file__)
        path = os.path.join(folder, os.path.expanduser("~/openai.json"))
        with open(path, encoding="utf-8") as fd:
            openai.api_key = json.loads(fd.read())["api_key"]
    except Exception as e: # pylint: disable=broad-except
        print(e)


METAPROMPT = '''
You are an expert Python Data Scientist. You know about Pandas, Matplotlib, Numpy, etc.
When your answer is a list of things, return a Python list.
when your answer looks like a table, return a Python dictionary.
Generate python code to be used in a Jupyter Notebook cell.
On the last line, just give the expression, do not call print(expression).
Replace calls to 'figure.show()' with 'figure'.
Do not generate explanations, but just show the Python code.
'''


def complete(prompt):
    """
    Generates a code completion using the currently selected language model.
    
    Args:
        prompt (str): The input prompt to generate the completion from.
    
    Returns:
        dict: A dictionary containing the generated text completion.
    """
    return complete_with_openai(prompt)
    # return complete_with_ollama(prompt)


def complete_with_ollama(prompt): # pylint: disable=unused-argument
    """
    Generates a code completion using Ollama with the llama3 language model.
    Installation:
     - Install ollama from https://ollama.com/download
     - Run the installer
     - In the shell that pops up, enter "ollama run codellama:python"
    
    Args:
        prompt (str): The input prompt to generate the completion from.
    
    Returns:
        dict: A dictionary containing the generated text completion.
    """
    response = requests.post("http://127.0.0.1:11434/api/generate", json={
        "model": "gemma2:2b",
        "prompt": prompt
    }, timeout=5)
    response.raise_for_status()
    for line in response.iter_lines():
        body = json.loads(line)
        print("########", body)
    return {
        "status": "OK",
        "text": response.text,
    }


def complete_with_openai(prompt):
    """
    Generates a code completion using the OpenAI GPT-3.5 language model.
    
    Args:
        prompt (str): The input prompt to generate the completion from.
    
    Returns:
        dict: A dictionary containing the generated text completion.
    """
    model="gpt-3.5-turbo-instruct"
    load_key()
    if not openai.api_key:
        return {
            "text": MISSING_KEY,
        }
    return openai.Completion.create(
        model=model,
        prompt=f"{METAPROMPT}\n{prompt}.",
        temperature=0,
        max_tokens=1000,
        top_p=1.0,
        frequency_penalty=0.0,
        presence_penalty=0.0,
        stop=["\"\"\""]
    )["choices"][0]



"""
CopyRight (c) 2024 - Chris Laffra - All Rights Reserved.

This module contains unit tests for the Language Server Protocol (LSP) implementation in the `lsp` module.
"""

import sys
import unittest
import unittest.mock

sys.path.append("..")

from tests import mocks # pylint: disable=wrong-import-position,unused-import
from static import lsp # pylint: disable=wrong-import-position
from static import worker # pylint: disable=wrong-import-position

orig_fuzzy_parse = lsp.fuzzy_parse


class TestCompletePython(unittest.TestCase):
    """
    This class contains unit tests for the Language Server Protocol (LSP) implementation in
    the `lsp` module. The tests cover various aspects of the LSP completion functionality.
    """

    def set_text(self, text):
        """
        Helper function to set the text, line, and position for the current test case.
        """
        lines = text.split("\n")
        line = len(lines) - 1
        pos = len(lines[-1])
        return text, line, pos

    def test_completes_attributes(self):
        """
        Tests completion of attributes of imported modules.
        """
        text, line, pos = self.set_text("=\nimport math\nmath.s")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn("sin(x:Any)", completions)
        self.assertIn("sqrt(x:Any)", completions)

    def test_completes_imported_modules(self):
        """
        Tests completion of names of imported modules.
        """
        text, line, pos = self.set_text("=\nimport math\nma")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn("math", completions)

    def test_completes_variables_in_scope(self):
        """
        Tests completion of variables available in the current scope.
        """
        text, line, pos = self.set_text("=\nx1 = x2 = 10\nx")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn("x1", completions)
        self.assertIn("x2", completions)

    def test_function_in_scope(self):
        """
        Tests completion of function names available in the current scope.
        """
        text, line, pos = self.set_text("=\ndef function(): pass\nf")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn("function()", completions)

    def test_function_attributes(self):
        """
        Tests completion of function attributes available in the current scope.
        """
        text, line, pos = self.set_text("=\ndef function(): pass\nfunction.")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn("__name__", completions)

    def test_match_cap(self):
        """
        Tests completion of specific string methods available on a string.
        
        This test case checks that the code completion of "ca" on a string
        includes `capitalize()`, `isdecimal()`, and `casefold()`.
        """
        text, line, pos = self.set_text("=\nx = 'hello'\nx.ca")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn("capitalize()", completions)
        self.assertIn("isdecimal()", completions)
        self.assertIn("casefold()", completions)

    def test_cache_list(self):
        """
        Tests completion of list methods available on a list stored in the worker cache.
        """
        worker.cache["D13"] = []
        text, line, pos = self.set_text("=\nD13.")
        completions = lsp.complete_python(text, line, pos, worker.cache, {})
        self.assertIn("append(object:Any)", completions)
        self.assertIn("insert(index:Any, object:Any)", completions)
        self.assertIn("clear()", completions)

    def test_cache_dict(self):
        """
        Tests completion of dictionary keys available in the current scope.
        """
        worker.cache["D14"] = { "dogs": 0, "cats": 1 }
        text, line, pos = self.set_text("=\nD14[")
        completions = lsp.complete_python(text, line, pos, worker.cache, {})
        self.assertIn('["cats"]', completions)
        self.assertIn('["dogs"]', completions)

    def test_dict_assign(self):
        """
        Tests completion of dictionary keys available in the current scope.
        """
        text, line, pos = self.set_text("=\nmy_dict = { 'dogs': 0, 'cats': 1 }\nmy")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn('my_dict', completions)

    def test_str_isdecimal(self):
        """
        Tests completion of the `isdecimal()` string method on a string.
        """
        text, line, pos = self.set_text("=\ns='abc'\ns.isd")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn('isdecimal()', completions)
        self.assertNotIn('capitalize()', completions)

    def test_sorting(self):
        """
        Tests completion of the `center()` string method on a string.
        """
        text, line, pos = self.set_text("=\ns='abc'\ns.ce")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertEqual("center(width:Any, fillchar:Any)", completions[0])

    def test_if(self):
        """
        Tests completion of the `center()` string method on a string.
        """
        text, line, pos = self.set_text("=\ns='abc'\nif s.")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn("center(width:Any, fillchar:Any)", completions)

    def test_for(self):
        """
        Tests completion of variables available in the current scope when using a for loop.
        """
        text, line, pos = self.set_text("=\nstring1='abc'\nstring2='def'\nfor s")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn("string1", completions)
        self.assertIn("string2", completions)

    def test_for_in(self):
        """
        Tests completion of variables available in the current scope when using a for loop.
        """
        text, line, pos = self.set_text("=\nstring1='abc'\nstring2='def'\nfor s in s")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn("string1", completions)
        self.assertIn("string2", completions)

    def test_dict_subscript(self):
        """
        Tests completion of dictionary keys available in the current scope.
        """
        text, line, pos = self.set_text("=\nmy_dict = { 'dogs': 0, 'cats': 1 }\nmy_dict[")
        completions = lsp.complete_python(text, line, pos, {}, {})
        self.assertIn('["dogs"]', completions)
        self.assertIn('["cats"]', completions)



"""
Copyright (c) 2024 laffra - All Rights Reserved. 

Represents a cell view in the sheet, which manages the display and behavior of a single cell.

"""

import ltk
import ltk.pubsub

import api
import constants
import history
import models
import preview
import selection
import state


class CycleError(Exception):
    """
    Represents a circular dependency in the cell calculation graph.
    """



class CellView(ltk.Widget): # pylint: disable=too-many-public-methods
    """
    Represents a cell view in the sheet, managing the display and behavior of a single cell.
    """

    observer = ltk.window.MutationObserver.new(ltk.proxy(lambda records, _: CellView.cellview_mutated(records))) # pylint: disable=unnecessary-lambda
    observer_config = ltk.to_js({
        "childList": True,
        "characterData": True,
        "attributes": True,
        "subtree": True
    })
    sheet = None

    def __init__(self, sheet, key: str, model: models.Cell, td=None):
        super().__init__()
        CellView.sheet= sheet
        self.model = model
        if not key:
            raise ValueError("Missing key for cell")
        if not self.model:
            raise ValueError(f"No model for cell {key}")
        self.element = td or ltk.find(f"#{self.model.key}")
        if not self.element.length:
            ltk.window.fillSheet(model.column, model.row)
            self.element = ltk.find(f"#{self.model.key}")
        self.running = False
        self.needs_worker = False
        self.inputs = set()
        self.dependents = set()
        if self.model.script != self.model.value:
            self.set(self.model.script, evaluate=False)
        self.on("mouseenter", ltk.proxy(lambda event: self.enter()))
        self.model.listen(self.model_changed)
        self.observer.observe(self.element[0], self.observer_config)

    @classmethod
    def cellview_mutated(cls, mutation_records):
        """
        One or more cellviews were mutated.
        """
        for key in set(record.target.id for record in mutation_records):
            if key and key in cls.sheet.cell_views:
                cls.sheet.cell_views[key].ui_changed()

    def ui_changed(self):
        """
        Updates the cell's value and script based on changes made in the UI.
        """
        new_value = str(self.element.attr("worker-set"))
        if new_value not in ["None", "<undefined>"]:
            self.set(new_value)
            self.element.removeAttr("worker-set")

    def model_changed(self, model, info):
        """
        Handles updates to the cell model, updating the cell's UI.

        Args:
            model (models.Cell): The cell model that was updated.
            info (dict): A dictionary containing information about the update,
            including the name of the property that changed.
        """
        if self.sheet.freeze_notifications:
            return
        if info["name"] == "script":
            self.set(model.script)
        elif info["name"] == "value":
            self.text(model.value)
        elif info["name"] == "style":
            self.css(model.style)
        ltk.schedule(self.sheet.run_ai, "run ai", 3)

    def enter(self):
        """
        Draws cell arrows and raises a preview when the cell is entered.
        """
        selection.remove_arrows(0)
        self.draw_cell_arrows()
        cell_preview = ltk.find(f"#preview-{self.model.key}")
        cell_preview.appendTo(cell_preview.parent()) # raise the preview to the top

    def set(self, script, evaluate=True):
        """
        Sets the script of the cell and evaluates it if necessary.

        Args:
            script (str): The new script to set for the cell.
            evaluate (bool, optional): When false does evaluate the new script.
        """
        self.remove_preview()
        if self.model.script != script:
            history.add(
                models.CellScriptChanged(self.model.key, self.model.script, script)
                    .apply(self.sheet.model)
            )
            self.model.script = script
        if self.sheet.current == self:
            self.sheet.editor.set(self.model.script)
            self.sheet.select(self)
        if not self.is_formula():
            self.sheet.cache[self.model.key] = api.convert(script)
        if evaluate:
            ltk.schedule(self.evaluate, f"eval-{self.model.key}")

    def is_running(self):
        """
        Determines whether the cell is currently running.
        """
        return self.running

    def start_running(self):
        """
        Starts the cell's evaluation.
        """
        self.running = True
        self.sheet.start_running(self)

    def stop_running(self):
        """
        Stops the cell's evaluation.
        """
        self.running = False
        self.sheet.stop_running(self)

    def is_formula(self):
        """
        Determines whether the cell's script is a formula.

        Returns:
            bool: True if the cell's script is a formula, False otherwise.
        """
        return isinstance(self.model.script, str) and self.model.script.startswith("=")

    def update(self, duration, value):
        """
        Updates the cell's text and value, and notifies any dependent cells.

        Args:
            duration (float): The duration in seconds of the last evaluation of the cell's script.
            value (Any): The result of evaluating the cell's script.
        """
        if self.model.script:
            count = self.sheet.counts[self.model.key]
            if count:
                if isinstance(value, str) and not "Error" in value:
                    if count > 1:
                        message = f"Ran in worker {count} times, last run took {duration:.3f}s"
                    else:
                        message = f"Ran once in worker, taking {duration:.3f}s"
                    state.console.write(
                        self.model.key,
                        f"[DAG] {self.model.key}: {message}"
                    )
        self.text(str(value))
        if value not in [self.model.value, self.model.script]:
            history.add(
                models.CellValueChanged(self.model.key, self.model.value, value)
                    .apply(self.sheet.model
                )
            )
        self.notify()
        self.model.value = value
        if self.sheet.current == self:
            self.sheet.selection.val(self.text())
        self.sheet.multi_selection.draw()

    def get_preview(self, value):
        """
        Returns a formatted HTML representation of the provided dict.

        Args:
            value (dict): The object to be converted to html.

        Returns:
            str: An HTML string.
        """
        return api.get_dict_table(value) if isinstance(value, dict) else None


    def notify(self):
        """
        Notifies any dependent cells of changes to this cell.

        This method is called when the cell's value or other properties have been updated,
        in order to trigger re-evaluation of any cells that depend on this cell's value.
        """
        for key in self.dependents:
            self.sheet.get_cell(key).evaluate()

    def worker_ready(self):
        """
        Marks the worker as ready and evaluates the cell.

        This method is called when the worker has finished executing the cell's script.
        It removes the cell from the sheet's cache, sets the running flag to False,
        and then evaluates the cell.
        """
        if self.model.key in self.sheet.cache:
            del self.sheet.cache[self.model.key]
        self.stop_running()
        self.evaluate()

    def select(self):
        """
        Selects the current cell, removes any existing cell arrows, saves the current
        selection position, sets the cell script in the editor, updates the AI prompt
        input, sets the selection text, shows the cell attributes container, sets the
        CSS editors, and scrolls the selection into view.
        """
        self.remove_arrows()
        self.sheet.editor.set(self.model.script)
        ltk.find("#selection").text(f"Cell: {self.model.key}")
        ltk.find("#cell-attributes-container").css("display", "block")
        self.set_prompt()
        self.set_css_editors()
        selection.scroll(self)

    def set_prompt(self):
        """
        Sets the cell's prompt.
        """
        prompt_editor = ltk.find("#ai-prompt")
        if not prompt_editor.hasFocus():
            prompt_editor.val(self.model.prompt)

    def set_css_editors(self):
        """
        Sets the CSS editor values for the current cell based on the cell's current CSS styles.
        """
        font_family = self.css("font-family")
        ltk.find("#cell-font-family").val(font_family or constants.DEFAULT_FONT_FAMILY)

        font_size = round(ltk.window.parseFloat(self.css("font-size")))
        ltk.find("#cell-font-size").val(font_size or constants.DEFAULT_FONT_SIZE)

        vertical_align = self.css("vertical-align")
        ltk.find("#cell-vertical-align").val(vertical_align or constants.DEFAULT_VERTICAL_ALIGN)

        text_align = self.css("text-align").replace("start", "left")
        ltk.find("#cell-text-align").val(text_align or constants.DEFAULT_TEXT_ALIGN)

        font_style = self.css("font-style")
        ltk.find("#cell-font-style").val(font_style or constants.DEFAULT_FONT_STYLE)

        font_weight = {"400": "normal", "700": "bold"}[self.css("font-weight")]
        ltk.find("#cell-font-weight").val(font_weight or constants.DEFAULT_FONT_WEIGHT)

        color = api.rgb_to_hex(self.css("color")) or constants.DEFAULT_COLOR
        ltk.find("#cell-color").val(color).css("background", color)

        background = api.rgb_to_hex(self.css("background-color")) or constants.DEFAULT_FILL
        ltk.find("#cell-fill").val(background).css("background", background)

    def clear(self):
        """
        Clears the current cell by:
        - Resetting the cell's CSS styles to default values
        - Removing any preview and completion elements associated with the cell
        - Clearing the cell's text and model data
        - Removing the cell from the sheet's cache and cells dictionary
        - Adding history events for the changes made to the cell
        - Notifying the sheet that the cell has been cleared
        - Reselecting the sheet to update the UI
        """
        self.css({
            "font-family": constants.DEFAULT_FONT_FAMILY,
            "font-size": constants.DEFAULT_FONT_SIZE,
            "font-style": constants.DEFAULT_FONT_STYLE,
            "color": constants.DEFAULT_COLOR,
            "background-color": constants.DEFAULT_FILL,
            "vertical-align": constants.DEFAULT_VERTICAL_ALIGN,
            "font-weight": constants.DEFAULT_FONT_WEIGHT,
            "text-align": constants.DEFAULT_TEXT_ALIGN,
        })

        ltk.find(f"#completion-{self.model.key}").remove()
        state.console.remove(f"ai-{self.model.key}")

        self.text("")
        self.model.clear(self.sheet.model)
        self.inputs.clear()
        if self.model.key in self.sheet.cell_views:
            del self.sheet.cell_views[self.model.key]
        self.sheet.cache[self.model.key] = 0

        history.add(models.CellScriptChanged(key=self.model.key, script=""))
        history.add(models.CellValueChanged(key=self.model.key, value=""))
        history.add(models.CellStyleChanged(key=self.model.key, style={}))
        self.notify()

        self.sheet.reselect()

    def remove_preview(self):
        """
        Removes the preview element associated with the cell and removes the cell from the preview collection.
        """
        ltk.find(f"#preview-{self.model.key}").remove()
        preview.remove(self.model.key)

    def draw_cell_arrows(self):
        """
        Draws the arrows indicating dependencies between cells on the sheet.
        """
        self.draw_arrows([])
        self.adjust_arrows()

    def remove_arrows(self):
        """
        Removes any arrow markers that have been drawn on the sheet.
        """
        selection.remove_arrows()

    def report_cycle(self, seen):
        """
        Detects and reports a dependency cycle in the cell's inputs.
        
        Args:
            seen (list): A list of cell keys that have been visited so far in the dependency graph.
        
        This function is called when a dependency cycle is detected while drawing the cell's
        input arrows.  It appends the current cell's key to the `seen` list,
        constructs a string representation of the cycle, and writes an error message to the console.
        """
        seen.append(self.model.key)
        cycle = " ⬅ ".join(seen)
        state.console.write(
            self.model.key,
            f"[ERROR] {self.model.key}: Dependency cycle detected: {cycle}"
        )

    def draw_arrows(self, seen):
        """
        Draws the arrows indicating dependencies between this cell and other cells on the sheet.
        """
        if self.model.key in seen:
            self.report_cycle(seen)
            return
        seen.append(self.model.key)
        self.remove_arrows()
        if state.mobile():
            return
        if not self.inputs:
            return
        cells = [ self.sheet.get_cell(input) for input in self.inputs ]
        ltk.window.addArrow(self.create_marker(cells, "inputs-marker arrow", seen), self.element)
        self.addClass("arrow")

    def create_marker(self, cells, clazz, seen):
        """
        Creates a marker element that represents a group of cells on the sheet.
        
        The marker is a div element that is positioned to encompass the cells, and has a class
        name that can be used to style it. The marker also has an event handler attached that
        removes any arrow markers when the user moves the mouse over it.
        
        Args:
            cells (list): A list of cell objects that the marker represents.
            clazz (str): A CSS class name to apply to the marker element.
            seen (list): A list of cell keys that have been visited so far in the dependency graph.
        
        Returns:
            The created marker element.
        """
        if not cells:
            return None
        if len(cells) == 1:
            cells[0].draw_arrows(seen)
            return cells[0]
        top, left, bottom, right = 10000, 10000, 0, 0
        for cell in cells:
            position = cell.position()
            left = min(position.left, left)
            top = min(position.top, top)
            right = max(position.left + cell.outerWidth(), right)
            bottom = max(position.top + cell.outerHeight(), bottom)
        return (ltk.Div()
            .addClass("marker")
            .addClass(clazz)
            .css("left", left)
            .css("top", top)
            .width(round(right - left - 4))
            .height(round(bottom - top - 5))
            .on("mousemove", ltk.proxy(lambda event: selection.remove_arrows()))
            .appendTo(ltk.find(".sheet-grid"))
        )

    def adjust_arrows(self):
        """
        Adjusts the position of arrow lines on the sheet to account for scrolling.
        """
        ltk.find(".leader-line").appendTo(ltk.find("#sheet-scrollable"))
        container = ltk.find("#sheet-container")
        scroll_left = container.scrollLeft()
        scroll_top = container.scrollTop()
        for arrow_line in ltk.find_list(".leader-line"):
            arrow_line \
                .css("top", ltk.window.parseFloat(arrow_line.css("top")) + scroll_top - 49) \
                .css("left", ltk.window.parseFloat(arrow_line.css("left")) + scroll_left)

    def edited(self, script):
        """
        Updates the cell's value with the provided script when the user edited the cell.
        
        Args:
            script (str): The new script to set for the cell.
        """
        self.set(script)

    def get_input_cells(self):
        """
        Returns a dictionary of input cell values for the current cell.
        
        The dictionary maps the input cell keys to their current values from the sheet cache.
        This is used to provide the input values when evaluating the cell's formula or script.
        """
        return {
            key: self.sheet.cache.get(key, 0)
            for key in self.inputs
        }

    def evaluate(self):
        """
        Evaluates the cell's value based on its script or formula.
        
        If the cell represents a formula, it resolves the input cells required to evaluate it.
        Otherwise, it updates the cell's value with the provided script.
        """
        state.console.remove(self.model.key)
        if self.is_formula():
            self.resolve_inputs()
        else:
            self.update(0, self.model.script)

    def show_loading(self):
        """
        Shows a loading indicator on the cell if the worker version is set to WORKER_LOADING.
        """
        if state.WORKER_VERSION != constants.WORKER_LOADING:
            return
        text = self.text()
        if not text.startswith(constants.ICON_HOUR_GLASS):
            self.text(f"{constants.ICON_HOUR_GLASS} {text}")

    def resolve_inputs(self):
        """
        Resolves the input cells required to evaluate the current cell's formula or script.
        """
        if self.is_running():
            return
        self.start_running()
        ltk.publish(
            "Application",
            "Worker",
            constants.TOPIC_WORKER_FIND_INPUTS,
            {
                "key": self.model.key, 
                "script": self.model.script[1:]
            },
        )
        # result will arrive in handle_inputs

    def inputs_missing(self):
        """
        Checks if any of the input cells required to evaluate the current cell are missing.
        
        Returns:
            bool: True if any of the input cells are missing, False otherwise.
        """
        for key in self.inputs:
            cell = self.sheet.get_cell(key)
            if cell.is_formula() and not self.sheet.counts[key]:
                return True
        return False

    def set_inputs(self, inputs):
        """
        Sets the input cells for the current cell.
        
        This method is typically called after resolving the input cells
        required to evaluate the current cell.
        """
        self.running = False
        self.inputs = inputs
        for key in self.get_input_cells():
            cell = self.sheet.get_cell(key)
            cell.dependents.add(self.model.key)

    def handle_inputs(self, inputs):
        """
        Handles the result of resolving the input cells required to evaluate
        the current cell's formula or script.
        
        This method is called after the worker has found the input cells needed to
        evaluate the current cell.  It sets the input cells, checks if any are missing,
        and then publishes a message to the worker to run the cell's script.
        
        If any of the input cells are missing, this method will return without publishing
        the message to the worker. The cell will remain in a "loading" state until the
        missing inputs are resolved.
        
        If all input cells are present, this method will increment the cell's execution
        count, mark the cell as needing worker processing, and publish a message to the
        worker to run the cell's script.
        """
        self.stop_running()
        if self.model.key in inputs:
            self.report_cycle(inputs)
        self.set_inputs(inputs)
        if self.inputs_missing():
            return
        self.sheet.counts[self.model.key] += 1
        self.needs_worker = True
        self.show_loading()
        ltk.publish(
            "Application",
            "Worker",
            ltk.TOPIC_WORKER_RUN,
            [self.model.key, self.model.script[1:], self.get_input_cells()]
        )
        # result will arrive in handle_worker_result

    def handle_worker_result(self, result):
        """
        Handles the result of a worker job that was run to evaluate the current cell's script.
        
        This method is called after the worker has finished running the cell's script. It processes the
        result of the worker job, updating the cell's value and state accordingly.
        
        If the worker job encountered an error, this method will handle the error, displaying it in the
        console and marking the line in the editor where the error occurred.
        
        If the worker job completed successfully, this method will update the cell's value and notify
        any dependents of the cell that the value has changed.
        """
        self.stop_running()
        if self.model.script == "":
            self.update(0, "")
            return
        if result["error"]:
            error = result["error"]
            duration = result["duration"]
            lineno = result["lineno"]
            tb = result["traceback"]
            parts = error.split("'")
            if len(parts) == 3 and parts[0] == "name " and parts[2] == " is not defined":
                key = parts[1]
                if key in self.inputs:
                    # The worker job ran out of sequence, ignore this error for now
                    return
            self.update(duration, error)
            if self.sheet.current.model.key == self.model.key:
                self.sheet.editor.mark_line(lineno)
            last_tb_lines = "\n".join(tb.split("\n")[-2:])
            state.console.write(self.model.key, f"[Error] {self.model.key}: Line {lineno}: {last_tb_lines}")
            ltk.window.console.orig_log(tb)
            return
        value = result["value"]
        if isinstance(value, str):
            value = value[1:-1] if value.startswith("'") and value.endswith("'") else value
        self.update(result["duration"], value)
        self.notify()

    def __repr__(self):
        return f"cell[{self.model.key}]"



"""
CopyRight (c) 2024 - Chris Laffra - All Rights Reserved.

Represents a multi-selection of cells in a sheet. Provides methods for selecting,
copying, pasting, and modifying the selected cells.
"""

import json

import api
import constants
import history
import ltk
import models
import state


class MultiSelection(): # pylint: disable=too-many-instance-attributes
    """
    Represents a multi-selection of cells in a sheet. Provides methods for selecting,
    copying, pasting, and modifying the selected cells.
    """

    def __init__(self, sheet):
        self.sheet = sheet
        self.started = False
        self.line_top = ltk.Div().addClass("multi-select top")
        self.line_left = ltk.Div().addClass("multi-select left")
        self.line_bottom = ltk.Div().addClass("multi-select bottom")
        self.line_right = ltk.Div().addClass("multi-select right")
        self.dimensions = [0, 0, 0, 0]
        self.cells = []
        self.cell1 = self.cell2 = None

        self.handler_by_shortcut = {
            "a": self.select_all,
            "b": self.bold,
            "c": self.copy,
            "i": self.italicize,
            "v": self.paste,
            "x": self.cut,
            "z": self.undo
        }

    def skip(self, event):
        """
        Skips the current event.
        
        This method is a placeholder for event handlers that are not implemented. It does nothing and
        simply returns without performing any actions.
        """

    def handle(self, event):
        """
        Handles keyboard shortcuts for the MultiSelection class. 
        """
        self.handler_by_shortcut.get(event.key, self.skip)(event)

    def select(self, cell):
        """
        Selects a single cell in the sheet and sets it as the start and end of the multi-selection.
        
        Args:
            cell (Cell): The cell to select.
        """
        self.cell1 = self.cell2 = cell
        self.stop(cell)

    def draw(self):
        """
            Draws the multi-selection overlay on the sheet grid. The overlay is composed of four divs
            representing the top, left, bottom, and right borders of the selected area. The positions
            and dimensions of these divs are calculated based on the positions of the start
            and end cells of the multi-selection.
        """
        ltk.find(".multi-select").remove()
        if not self.cell1 or not self.cell2 or self.cell1 is self.cell2:
            return
        pos1, pos2 = self.cell1.position(), self.cell2.position()
        if not pos1 or not pos2:
            return
        left = min(pos1.left, pos2.left)
        top = min(pos1.top, pos2.top) - 1
        right = max(pos1.left + self.cell1.outerWidth(), pos2.left + self.cell2.outerWidth()) - 1
        bottom = max(pos1.top + self.cell1.outerHeight(), pos2.top + self.cell2.outerHeight()) - 2
        ltk.find(".sheet-grid").append(
            self.line_top.css("left", left).css("top", top).css("width", right - left),
            self.line_left.css("left", left).css("top", top).css("height", bottom - top),
            self.line_bottom.css("left", left).css("top", bottom).css("width", right - left + 2),
            self.line_right.css("left", right).css("top", top).css("height", bottom - top),
        )

    def set_dimensions(self):
        """
        Sets the dimensions of the multi-selection based on the positions of the start and end cells.
        
        The dimensions are calculated as follows:
        - The left column is the minimum of the start and end cell column indices.
        - The right column is the maximum of the start and end cell column indices.
        - The top row is the minimum of the start and end cell row indices.
        - The bottom row is the maximum of the start and end cell row indices.
        
        The dimensions are stored in the `self.dimensions` attribute as a list in the format
        `[from_col, to_col, from_row, to_row]`.
        """
        pos1, pos2 = self.cell1.position(), self.cell2.position()
        assert self.cell1.model.column, f"Column should be >= 1, not {self.cell1.model.column} for {self.cell1}"
        self.dimensions = [
            self.cell1.model.column if pos1.left < pos2.left else self.cell2.model.column,
            self.cell2.model.column if pos1.left < pos2.left else self.cell1.model.column,
            self.cell1.model.row if pos1.top < pos2.top else self.cell2.model.row,
            self.cell2.model.row if pos1.top < pos2.top else self.cell1.model.row,
        ]

    def highlight_col_row(self):
        """
        Highlights the column and row range specified by the `self.dimensions` attribute.
        """
        ltk.window.highlightColRow(*self.dimensions)


    def css(self, name, value):
        """
        Sets the CSS style of the selected cells in the sheet.
        
        Args:
            name (str): The name of the CSS property to set.
            value (str): The value to set for the CSS property.
        """
        keys = ",".join([f"#{cell}" for cell in self.cells])
        ltk.find(keys).css(name, value)
        self.sheet.selection.css(name, value)
        with history.SingleEdit(f"Update '{name}' to '{value}' for {','.join(self.cells)}"):
            for model in [self.sheet.get_cell(key).model for key in self.cells]:
                _style = model.style.copy()
                model.style[name] = value
                history.add(models.CellStyleChanged(model.key, _style, model.style))
        self.draw()

    def undo(self, event):
        """
        Undoes the last action performed on the sheet.
        """
        history.undo(self.sheet.model)
        event.preventDefault()

    def bold(self, event): # pylint: disable=unused-argument
        """
        Toggles the font weight of the first selected cell between normal (400) and bold (700).
        """
        if not self.cells:
            return
        cell = self.sheet.get_cell(self.cells[0])
        self.css("font-weight", {"400": "700", "700": "400"}[cell.css("font-weight")])

    def italicize(self, event): # pylint: disable=unused-argument
        """
        Toggles the font style of the first selected cell between normal and italic.
        """
        if not self.cells:
            return
        cell = self.sheet.get_cell(self.cells[0])
        self.css("font-style", {"normal":"italic", "italic":"normal"}[cell.css("font-style")])

    def copy(self, event): # pylint: disable=unused-argument
        """
        Copies the selected cells in the sheet to the clipboard as both text and HTML.
        """
        from_col, to_col, from_row, to_row = self.dimensions

        def get_cell(col, row):
            return self.sheet.get_cell(api.get_key_from_col_row(col, row))

        def get_script(col, row):
            return get_cell(col, row).model.script

        def get_td(col, row):
            cell = get_cell(col, row)
            return ltk.TableData(cell.model.script).attr("style", cell.attr("style"))

        text = "\n".join([
                "\t".join([
                    get_script(col, row)
                    for col in range(from_col, to_col + 1)
                ])
                for row in range(from_row, to_row + 1)
            ])

        html = ltk.Div(ltk.Table( *[
            ltk.TableRow(*[
                get_td(col, row)
                for col in range(from_col, to_col + 1)
            ])
            for row in range(from_row, to_row + 1)
        ])).html()

        ltk.window.clipboardWrite(text, html)

    def paste(self, event):
        """
        Pastes the contents of the clipboard into the current cell of the sheet.
        If the clipboard contains text, it is pasted as plain text.
        If the clipboard contains HTML, it is pasted as HTML.
        The pasted content is inserted at the current cell's position, and the sheet
        is reselected after the paste operation is complete.
        """
        current = self.sheet.current

        def paste_done(infos, text):
            start = ltk.get_time()
            def find_cells(infos):
                state.console.write("paste", f"[Paste] Loading {len(infos)} pasted cells")
                for key, new_script, new_style in infos[:constants.MAX_EDITS_PER_SYNC]:
                    cell_view = self.sheet.get_cell(key)
                    cell_view.set(new_script)
                    cell_model = cell_view.model
                    cell_model.style = json.loads(new_style)
                infos = infos[constants.MAX_EDITS_PER_SYNC:]
                if infos:
                    ltk.schedule(lambda: find_cells(infos), "handle more pasted cells")
                else:
                    state.console.write("paste",
                            f"[Paste] Pasting {len(text):,} bytes took {ltk.get_time() - start:.3f}s")
                self.sheet.select(self.sheet.current)

            infos = list(infos)
            with history.SingleEdit(f"Paste {len(infos)} cells"):
                find_cells(infos)
            self.sheet.reselect()

        def process_clipboard(text):
            state.console.write("paste", f"[Paste] Pasting {len(text):,} bytes from the clipboard...")
            def paste_text():
                ltk.window.pasteText(
                    text,
                    current.model.column,
                    current.model.row,
                    ltk.proxy(lambda keys: paste_done(keys, text))
                )
            def paste_html():
                ltk.window.pasteHTML(text,
                    current.model.column,
                    current.model.row,
                    ltk.proxy(lambda keys: paste_done(keys, text))
                )
            if event.shiftKey:
                paste_text()
            else:
                paste_html()

        state.console.write("paste", "[Paste] Pasting from the clipboard...")
        def get_clipboard():
            ltk.window.getClipboard(ltk.proxy(process_clipboard), not event.shiftKey)
        ltk.schedule(get_clipboard, "getting clipboard")

    def cut(self, event):
        """
        Cuts the selected cells from the sheet.
        """
        self.copy(event)
        self.clear()

    def clear(self):
        """
        Clears the selected cells in the sheet.
        """
        with history.SingleEdit(f"Clear {len(self.cells)} cells"):
            for key in self.cells:
                if key in self.sheet.model.cells:
                    self.sheet.get_cell(key).clear()
        self.draw()

    def start(self, cell):
        """
        Starts the selection process by focusing the main element, selecting the provided cell,
        setting the start and end cells to the provided cell, updating the selection, 
        and marking the selection as started.
        """
        ltk.find("#main").focus()
        self.sheet.select(cell)
        self.cell1 = self.cell2 = cell
        self.update()
        self.started = True

    def extend(self, cell, force=False):
        """
        Extends the current selection to the provided cell.
        """
        if not force and not self.started:
            return
        self.cell2 = cell
        self.update()
        self.highlight_col_row()

    def select_all(self, event):
        """
        Selects all cells in the sheet and updates the selection.
        """
        cells = self.sheet.cells.values()
        if cells:
            min_col = min(cell.column for cell in cells)
            max_col = max(cell.column for cell in cells)
            min_row = min(cell.row for cell in cells)
            max_row = max(cell.row for cell in cells)
            self.cell1 = self.sheet.get_cell(api.get_key_from_col_row(min_col, min_row))
            self.cell2 = self.sheet.get_cell(api.get_key_from_col_row(max_col, max_row))
            self.update()
        event.preventDefault()

    def stop(self, cell): # pylint: disable=unused-argument
        """
        Stops the selection process, updates the selection, and highlights the column and row.
        """
        self.started = False
        self.update()
        self.highlight_col_row()

    def update(self):
        """
        Updates the selection by setting the dimensions, getting the cell keys within 
        those dimensions, and drawing the selection.
        """
        self.set_dimensions()
        self.cells = self.sheet.model.get_cell_keys(*self.dimensions)
        self.draw()


def remove_arrows(delay=1000):
    """
    Removes any visible arrows or input markers from the UI after a delay of 1 second.
    """
    def dissolve_arrow(arrow):
        arrow.animate({ "opacity": 0 }, delay, ltk.proxy(lambda: arrow.remove())) # pylint: disable=unnecessary-lambda
    for arrow in ltk.find_list(".leader-line, .inputs-marker"):
        if arrow.css("opacity") == "1":
            dissolve_arrow(arrow)
    ltk.find(".arrow").removeClass("arrow")


def scroll(cell):
    """
        Schedules the `scroll_now` function to be executed after a short delay.
        This allows the scrolling to be performed asynchronously, which 
        improves the responsiveness of the user interface.
    """
    ltk.schedule(lambda: scroll_now(cell), "scroll later")


def scroll_now(cell):
    """
    Scrolls the sheet grid to ensure the given cell is visible within the sheet container.
    
    This function is responsible for calculating the necessary margin adjustments to the sheet grid
    in order to center the given cell within the visible area of the sheet container. It uses
    animation to smoothly transition the grid to the new position.
    
    Args:
        cell (object): The cell object that should be made visible.
    """
    container = ltk.find("#sheet-container")
    if not container.length:
        return

    grid = ltk.find(".sheet-grid")
    margin_left = ltk.window.parseInt(grid.css("margin-left"))
    margin_top = ltk.window.parseInt(grid.css("margin-top"))
    offset = cell.position()
    style = {}

    if offset.left + margin_left - 61 < 0:
        style["margin-left"] = -offset.left + 61
    elif offset.left + margin_left + 61 > container.width():
        style["margin-left"] = -offset.left + container.width() - cell.width() - 4
    if offset.top + margin_top - 30 < 0:
        style["margin-top"] = -offset.top + 30
    elif offset.top + margin_top + 30 > container.height():
        style["margin-top"] = -offset.top + container.height() - cell.height() - 4

    def step(*args): # pylint: disable=unused-argument
        ltk.schedule(lambda: ltk.window.adjustSheetPosition(), "adjust sheet position") # pylint: disable=unnecessary-lambda

    grid.animate(
        ltk.to_js(style),
        {
            "duration": constants.ANIMATION_DURATION_FAST,
            "step": ltk.proxy(step)
        }
    )



"""
Copyright (c) 2024 laffra - All Rights Reserved. 

Global constants used by the UI, worker, and models.
"""

DEFAULT_COLUMN_COUNT = 26
DEFAULT_ROW_COUNT = 65
BIG_SHEET_SIZE = 2000
LARGE_PREVIEW_SIZE = 3 * 1024 * 1024

DEFAULT_COLUMN_WIDTH = 72
DEFAULT_ROW_HEIGHT = 16
DEFAULT_FONT_FAMILY = "Arial"
DEFAULT_FONT_SIZE = "12px"
DEFAULT_COLOR = "rgb(0, 0, 0)"
DEFAULT_FILL = "rgb(255, 255, 255)"
DEFAULT_STYLE = "normal"
DEFAULT_TEXT_ALIGN = "left"
DEFAULT_VERTICAL_ALIGN = "bottom"
DEFAULT_FONT_STYLE = "normal"
DEFAULT_FONT_WEIGHT = "normal"

DEFAULT_STYLE = {
    "font-family": DEFAULT_FONT_FAMILY,
    "font-size": DEFAULT_FONT_SIZE,
    "font-style": DEFAULT_FONT_STYLE,
    "color": DEFAULT_COLOR,
    "background-color": DEFAULT_FILL,
    "vertical-align": DEFAULT_VERTICAL_ALIGN,
    "font-weight": DEFAULT_FONT_WEIGHT,
    "text-align": DEFAULT_TEXT_ALIGN,
}

ANIMATION_DURATION_FAST = 150
ANIMATION_DURATION_MEDIUM = 500
ANIMATION_DURATION_SLOW = 1000
ANIMATION_DURATION_VERY_SLOW = 3000
ANIMATION_DURATION = ANIMATION_DURATION_FAST
MAX_EDITS_PER_SYNC = 500

ICON_HOUR_GLASS = "⏳"
ICON_DATAFRAME = "🐼"
ICON_STAR = "⭐"
ICON_JSON = "🔲"
ICON_PLOT = "📊"
ICON_LIST = "↕"

TILE_SIZE = 10

OTHER_EDITOR_TIMEOUT = 60

MODE_PRODUCTION = "prod"
MODE_DEVELOPMENT = "dev"
WORKER_LOADING = "Loading..."

PREVIEW_HEADER_WIDTH = 56
PREVIEW_HEADER_HEIGHT = 32

IMAGE_COLORS = [ "#3B71CA", "#9FA6B2", "#14A44D", "#DC4C64", "#E4A11B", "#54B4D3", "#FBFBFB", "#332D2D" ]

FONT_NAMES = [ "Arial", "Courier", "Roboto" ]
FONT_SIZES = list(range(6, 21)) + list(range(24, 73, 4))

TOPIC_WORKER_PRINT = "worker.print"
TOPIC_WORKER_COMPLETION = "worker.completion"
TOPIC_WORKER_COMPLETE = "worker.complete"
TOPIC_WORKER_CODE_COMPLETE = "worker.code.complete"
TOPIC_WORKER_CODE_COMPLETION = "worker.code.completion"
TOPIC_WORKER_FIND_INPUTS = "worker.find.inputs"
TOPIC_WORKER_INPUTS = "worker.inputs"
TOPIC_API_SET_CELLS = "api.set_cells"

TOPIC_WORKER_PREVIEW_IMPORT_WEB = "worker.preview.import.web"
TOPIC_WORKER_PREVIEW_IMPORTED_WEB = "worker.preview.imported.web"
TOPIC_WORKER_IMPORT_WEB = "worker.import.web"
TOPIC_WORKER_IMPORTED_WEB = "worker.imported.web"
TOPIC_WORKER_UPLOAD = "worker.import.upload"
TOPIC_WORKER_UPLOADED = "worker.imported.uploaded"

URL = "url"
PROMPT = "prompt"
PYTHON_RUNTIME = "runtime"
STATUS = "status"
SHEET_ID = "id"
ERROR = "error"

PUBSUB_STATE_ID = "State"
PUBSUB_SHEET_ID = "Application"
